
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <ctype.h>
#include <X11/keysym.h>

#include "send/tclXtSend.h"

#include "funcs.h"

#define TM_NUM_PARAMS 32

static Tcl_Interp *interp;
static Boolean sleepDone;

static char lib_command[] = "if [file exists replay.rules] {source replay.rules}";
static char *lib_cmd[] = { 
	"if [file exists ", 
	"] {source ", 
	"}"
};
static char *replay_cmds;

Boolean RXt_verbose;

/*
 *--------------------------------------------------------------
 *
 * RXt_ExtractArg --
 *	find a name and its value in a list
 *
 * Results:
 *	returns a pointer to the value
 *
 * Side effects:
 *	none
 *--------------------------------------------------------------
 */

static char *
RXt_ExtractArg(name, argc, argv)
    char *name;
    int argc;
    char *argv[];
{
   int n;

    for (n = 0; n < argc; n++) {
	if (strcmp(name, argv[n]) == 0) {
	    if (++n < argc) {
		return argv[n];
	    } else {
		return NULL;
	    }
	}
    }
    return NULL;
}

/*
 *--------------------------------------------------------------
 *
 * RXt_MakeXEvent --
 *
 *	Create an X event that can be used to send to another window
 *
 * Results:
 *	a Tcl result value
 *
 * Side effects:
 *	returns a filled in event in xev
 *
 * Restrictions:
 *	only handles a small number of X types
 *--------------------------------------------------------------
 */
int
RXt_MakeXEvent(w, interp, xev, argc, argv)
    Widget w;
    Tcl_Interp *interp;
    XEvent *xev;
    int argc;
    char **argv;
{
    char *type;
    char *value;
    struct timeval  tp;
    Display *display;


    /* let them know we made this thing up */
    xev->xany.send_event = True;

    /* fill in the rest of the xany values.
     */
    xev->xany.display = display = XtDisplay(w);
    xev->xany.window = XtWindow(w);

    if ((type = RXt_ExtractArg("-type", argc, argv)) == NULL) {
        type = "ClientMessage";
    }

    if (type[0] == 'B') {
        if (strcmp(type, "ButtonRelease") == 0) {
	    xev->type = ButtonRelease;
	} else
	if (strcmp(type, "ButtonPress") == 0) {
	    xev->type = ButtonPress;
	} else {
	    sprintf(interp->result, "illegal type %s\n", type);
	    return TCL_ERROR;
	}

	/* set defaults */
	xev->xbutton.x = 0;
	xev->xbutton.y = 0;
	xev->xbutton.x_root = 0;
	xev->xbutton.y_root = 0;
	xev->xbutton.state = 0;
	xev->xbutton.button = Button1;
	xev->xbutton.same_screen = True;

	/* can't use clock() because we don't know timer resolution,
	   and lots of systems don't have CLK_TCK or CLOCKS_PER_SEC .
	   So do what Xt does:
	 */
        gettimeofday(&tp, 0);
        xev->xbutton.time = (tp.tv_sec * 1000) + (tp.tv_usec / 1000);

	if ((value = RXt_ExtractArg("-x", argc, argv)) != NULL) {
	    xev->xbutton.x = atoi(value);
	}

	if ((value = RXt_ExtractArg("-y", argc, argv)) != NULL) {
	    xev->xbutton.y = atoi(value);
	}

	if ((value = RXt_ExtractArg("-button", argc, argv)) != NULL) {
	    /* some form of pattern required here */
	    if (strcmp(value, "Button1") == 0)
		xev->xbutton.button = Button1;
	    else
	    if (strcmp(value, "Button2") == 0)
		xev->xbutton.button = Button2;
	    else
	    if (strcmp(value, "Button3") == 0)
		xev->xbutton.button = Button3;
	    else
	    if (strcmp(value, "Button4") == 0)
		xev->xbutton.button = Button4;
	    else
	    if (strcmp(value, "Button5") == 0)
		xev->xbutton.button = Button5;
	    else {
		sprintf(interp->result, "illegal button %s", value);
		return TCL_ERROR;
	    }
	}

	return TCL_OK;
    }

    if (type[0] == 'K') {
	unsigned int mask;
	KeySym keysym;
	KeyCode keycode;

        if (strcmp(type, "KeyRelease") == 0) {
	    xev->type = KeyRelease;
	} else
	if (strcmp(type, "KeyPress") == 0) {
	    xev->type = KeyPress;
	} else {
	    sprintf(interp->result, "illegal type %s\n", type);
	    return TCL_ERROR;;
	}

	/* can't use clock() because we don't know timer resolution,
	   and lots of systems don't have CLK_TCK or CLOCKS_PER_SEC .
	   So do what Xt does:
	 */
        gettimeofday(&tp, 0);
        xev->xkey.time = (tp.tv_sec * 1000) + (tp.tv_usec / 1000);

	if ((value = RXt_ExtractArg("-x", argc, argv)) != NULL) {
	    xev->xkey.x = atoi(value);
	} else {
	    xev->xkey.x = 0;
	}

	if ((value = RXt_ExtractArg("-y", argc, argv)) != NULL) {
	    xev->xkey.y = atoi(value);
	} else {
	    xev->xkey.y = 0;
	}

	if ((value = RXt_ExtractArg("-keysym", argc, argv)) != NULL) {
	    if ((keysym = XStringToKeysym(value)) == NoSymbol) {
		sprintf(interp->result, "illegal keysym\"%s\"", value);
		return TCL_ERROR;
	    }
	    if ((keycode = XKeysymToKeycode(display, keysym)) == 0) {
		sprintf(interp->result, "no keycode for keysym \"%s\"", value);
		return TCL_ERROR;
	    }

	    mask = 0;
	    /* is this what we should do? */
	    if (keysym >= XK_A && keysym <= XK_Z)
		mask |= ShiftMask;
	    
	    xev->xkey.keycode = keycode;
	    xev->xkey.state = mask;
	} else {
	    sprintf(interp->result, "Key event must have a keysym");
	    return TCL_ERROR;
	}

	return TCL_OK;
    }

    if (strcmp(type, "ClientMessage") == 0) {
	return TCL_OK;
    }

    sprintf(interp->result, "illegal type %s\n", type);
    return TCL_ERROR;;
}

/*
 *--------------------------------------------------------------
 *
 * RXt_EndToken --
 *
 *	return the next char past a string or quoted string
 *	using the actions parameter syntax (Asente & Swick appendix B)
 *
 * Results:
 *	return pointer to next char after string
 *
 *--------------------------------------------------------------
 */

char *
RXt_EndToken(s)
    char *s;
{
    if (s == NULL || *s == '\0')
	return s;

    /* quoted string */
    if (*s == '"') {
	s++;
	while (*s != '"' && *s != '\0')
	    s++;
	if (*s == '"')
	    return (s+1);
	else
	    return s;
    }

    /* unquoted string */
    while (*s != ' ' && *s != '\t' && *s != ',' &&
	   *s != '\n' && *s != ')' && *s != '\0')
	s++;
    return s;
}

/*
 *--------------------------------------------------------------
 *
 * RXt_ParseAction --
 *
 *	Parse an action into action name + args
 *
 * Results:
 *	sets the action, options and num_options
 *
 * Side-effects:
 * 	none
 *
 * Caveats:
 *	Somehow, this is an ugly piece of code. Tidy it up!
 *--------------------------------------------------------------
 */

int
RXt_ParseAction(orig, action, params, num_params)
    char *orig;
    char **action;
    char *params[];
    Cardinal *num_params;
{
    int n = 0;
    char *start, *end;
#   define skipblanks(p) while (isspace(*p)) p++;

    skipblanks(orig);
    /* make a new string, and put nulls in it at the end of tokens */
    start = XtNewString(orig);
    *action = end = start;

    /* delimit action name */
    while (isalnum(*end) || *end == '_' || *end == '-')
	end++;
    if (*end != '(') {
	fprintf(stderr, "no \"(\" after name\n");
	return TCL_ERROR;
    }
    *end = '\0';

    start = end + 1;
    for (n = 0; n < TM_NUM_PARAMS; True) {
	skipblanks(start);
	if (*start == ')')
	    break;

	/* new token */
	end = RXt_EndToken(start);

	/* lose quotes if it was a quoted string */
	if (*start == '"') {
	    start++;
	    *(end - 1) = '\0';
	}
	params[n++] = start;

	if (isspace(*end)) {
	    *end++ = '\0';
	}
	skipblanks(end);
	if (*end == '\0') {
	    fprintf(stderr, "unterminated action\n");
	    return TCL_ERROR;
	}
	if (*end == ')') {
	    *end = '\0';
	    break;
	}
	if (*end != ',') {
	    fprintf(stderr, "no \",\" delimiter\n");
	    return TCL_ERROR;
	}
	*end = '\0';
	start = end + 1;
    }

    *num_params = n;
    return TCL_OK;
}

static void
EventLoop(clientData, interp, level, command, cmdProc,
          cmdClientData, argc, argv)
    ClientData clientData;
    Tcl_Interp *interp;
    int level;
    char *command;
    Tcl_CmdProc *cmdProc;
    ClientData cmdClientData;
    int argc;
    char *argv[];
{
    XtAppContext appContext = (XtAppContext) clientData;

    while (XtAppPending(appContext)) {
	XtAppProcessEvent(appContext, XtIMAll);
    }
}

static void
EndSleep(clientData, timer)
    XtPointer clientData;
    XtIntervalId *timer;
{
    sleepDone = True;
}

static int
Sleep(clientData, interp, argc, argv)
    ClientData *clientData;
    Tcl_Interp *interp;
    int argc;
    char **argv;
{
    Widget w = (Widget) clientData;
    Display *display = XtDisplay(w);
    XtAppContext appContext = XtWidgetToApplicationContext(w);
    int timeOut;

    Tcl_GetInt(interp, argv[1], &timeOut);
    if (RXt_verbose)
	printf("sleeping for %d\n", timeOut);
    XtAppAddTimeOut(appContext, timeOut, EndSleep, NULL);
    sleepDone = False;

    /* we are guaranteed at least one event: when the timer goes off */
    while (1) {
	XtAppProcessEvent(appContext, XtIMAll);
	if (sleepDone)
	    return TCL_OK;
    }
}

static int
CallActionProc(clientData, interp, argc, argv)
    ClientData *clientData;
    Tcl_Interp *interp;
    int argc;
    char **argv;
{

        XEvent xev;
        char *action;
        char *params[TM_NUM_PARAMS];
        Cardinal num_params;
	Widget w;
	Widget toplevel;

        if (argc < 3) {
            sprintf(interp->result, "wrong # args: should be callActionProc \
widget action");
            return TCL_ERROR;
        }

	toplevel = (Widget) clientData;

	if ((w = XtNameToWidget(toplevel, argv[1])) == NULL) {
	    sprintf(interp->result, "no such widget: %s", argv[1]);
	    return TCL_ERROR;
	}

        if (RXt_MakeXEvent(w, interp, &xev, argc-3, argv+3) == TCL_ERROR) {
           return TCL_ERROR;
	}
        if (RXt_ParseAction(argv[2], &action, params, &num_params) != TCL_OK) {
            sprintf(interp->result, "parse error in action \"%.50s\"", argv[2]);
            return TCL_ERROR;
        }
	
	if (RXt_verbose) {
	    printf("action: %s on widget: ", action);
	    RXt_PrintFullName(w);
	    printf("\n");
	}
        XtCallActionProc(w, action, &xev, params, num_params);
        

    return TCL_OK;
}

static int
WarpPointer(clientData, interp, argc, argv)
    ClientData *clientData;
    Tcl_Interp *interp;
    int argc;
    char **argv;
{
    Widget toplevel = (Widget) clientData;
    Widget w;
    Window win;
    int x, y;

    if (argc < 4) {
        sprintf(interp->result, "wrong # args: should be warpPointer \
widget x y");
        return TCL_ERROR;
    }

    if ((w = XtNameToWidget(toplevel, argv[1])) == NULL) {
	sprintf(interp->result, "no window for %s", argv[1]);
	return TCL_ERROR;
    }

    if ((Tcl_GetInt(interp, argv[2], &x) == TCL_ERROR) ||
	(Tcl_GetInt(interp, argv[3], &y) == TCL_ERROR))
	return TCL_ERROR;

    win = XtWindowOfObject(w);

    XWarpPointer(XtDisplayOfObject(w), None, win, 0, 0, 0, 0, x, y);
    return TCL_OK;
}



static void
StartPlayer(clientData, timer)
    XtPointer clientData;
    XtIntervalId *timer;
{
    char *data_file = (char *) clientData;

    if (Tcl_VarEval(interp, lib_cmd[0], data_file, lib_cmd[1],
		data_file, lib_cmd[2], NULL) == TCL_ERROR) {
	fprintf(stderr, "%s\n", interp->result);
	exit(1);
    }
}

Bool
RXt_RegisterPlayer(toplevel, rules_file, data_file, verbose)
    Widget toplevel;
    char *rules_file;
    char *data_file;
    Boolean verbose;
{
    char *name;
    int fp;
    off_t size;
    struct stat stat_buf;
    
    RXt_verbose = verbose;

    interp = Tcl_CreateInterp();
    name = XtName(toplevel);

    RXt_RegisterConverters(interp, XtWidgetToApplicationContext(toplevel));

    Tcl_CreateCommand(interp, "callActionProc", CallActionProc,
		toplevel, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "getValues", RXt_GetValues,
		toplevel, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "sleep", Sleep,
		toplevel, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "warpPointer", WarpPointer,
		toplevel, (Tcl_CmdDeleteProc *) NULL);

    Tcl_VarEval(interp, lib_cmd[0], rules_file, lib_cmd[1],
		rules_file, lib_cmd[2], NULL);

    Tcl_CreateTrace(interp, 10000, EventLoop,
			XtWidgetToApplicationContext(toplevel));

    XtAppAddTimeOut(XtWidgetToApplicationContext(toplevel),
		0, StartPlayer, data_file);


    if (TclXtSend_RegisterInterp(interp, name, toplevel) == TCL_ERROR) {
	fprintf(stderr, "couldn't register interpreter %s\n", name);
	return False;
    }

    return True;
}
